VERSION = "2.0.0"
# Echonetlite message format:
#
# - EDT0     |Property value data             |01 ..|01 01 30 01
#  - NUM     |Total number of instances       |01   |1
#  - EOJ     |ECHONET Lite object specificat..|01 ..|01 30 01
#    - EOJX1 |Class group code                |01   |Air conditioner-related device class group
#    - EOJX2 |Class code                      |30   |Home air conditioner class
#    - EOJX3 |Instance code                   |01   |1

ENL_PORT = 3610
ENL_MULTICAST_ADDRESS = "224.0.23.0"

# ------------------------------------------------------------------
# EHD1: ECHONET Lite Header 1
# ---------------------------------------------------------------- */
EHD1 = {
    0x00: 'Not available',
    0x10: 'Conventional ECHONET Lite Specification'
}

# ------------------------------------------------------------------
# EHD1: ECHONET Lite Header 2
# ---------------------------------------------------------------- */
EHD2 = {
    0x81: 'Format 1 (specified message format)',
    0x82: 'Format 2 (arbitrary message format)'
}

# ------------------------------------------------------------------
# ESV
# ------------------------------------------------------------------
GETC = 			0x60
SETC = 			0x61
GET  = 			0x62
INFREQ =		0x63
SETGET = 		0x6E
SETRES =		0x71
GETRES =		0x72
INF =			0x73
INFC = 			0x74
INFC_RES =		0x7A
SETGET_RES =	0x7E
SETI_SNA = 		0x50
SETC_SND =		0x51
GET_SNA = 		0x52
INF_SNA = 		0x53
SETGET_SNA =	0x5E
INSTANCE_LIST = 0xD6

ESV_CODES = {
	0x60: {'name': 'GetC', 'description': 'Property value write request (no response required)'},
	0x61: {'name': 'SetC', 'description': 'Property value write request (response required)'},
	0x62: {'name': 'Get', 'description': 'Property value read request'},
	0x63: {'name': 'INF_REQ', 'description': 'Property value notification request'},
	0x6E: {'name': 'SetGet', 'description': 'Property value write & read request'},
	0x71: {'name': 'Set_Res', 'description': 'Property value Property value write response'},
	0x72: {'name': 'Get_Res' , 'description': 'Property value read response'},
	0x73: {'name': 'INF' , 'description': 'Property value notification'},
	0x74: {'name': 'INFC', 'description': 'Property value notification (response required)'},
	0x7A: {'name': 'INFC_Res' , 'description': 'Property value notification response'},
	0x7E: {'name': 'SetGet_Res' , 'description': 'Property value write & read response'},
	0x50: {'name': 'SetI_SNA', 'description': 'Property value write request (response not possible)'},
	0x51: {'name': 'SetC_SNA' , 'description': 'Property value write request (response not possible)'},
	0x52: {'name': 'Get_SNA', 'description': 'Property value read (response not possible)'},
	0x53: {'name': 'INF_SNA', 'description': 'Property value notification (response not possible)'},
    0x5E: {'name': 'SetGet_SNA', 'description': 'Property value write & read (response not possible)'}
}

ENL_ON = 0x30
ENL_OFF = 0x31
ENL_STATUS = 0x80
ENL_UID = 0x83
ENL_SETMAP = 0x9E
ENL_GETMAP = 0x9F



from pychonet.EchonetInstance import EchonetInstance

MEASURED_TEMP = 0xE0

class TemperatureSensor(EchonetInstance):
    def __init__(self, netif, instance = 0x1):
        self.eojgc = 0x00
        self.eojcc = 0x11
        EchonetInstance.__init__(self, self.eojgc, self.eojcc, instance, netif)

    def getMeasuredTemperature(self):
        return self.update(MEASURED_TEMP)



from datetime import datetime, timezone

#------------ EPC GENERIC FUNCTIONS -------
def _int(edt):
    return int.from_bytes(edt, 'big')

# Check status of Echonnet Instance
#----------------- EPC SUPER FUNCTIONS -----------------------------
def _0080(edt):
    ops_value = int.from_bytes(edt, 'big')
    return ('On' if ops_value == 0x30 else 'Off')

def _009X(edt):
    payload = []
    if len(edt) < 17:
        for i in range (1, len(edt)):
            payload.append(edt[i])
        return payload

    for i in range (1, len(edt)):
        code = i-1
        binary = '{0:08b}'.format(edt[i])[::-1]
        for j in range (0, 8):
            if binary[j] == "1":
                EPC = (j+8) * 0x10 + code
                payload.append(EPC)
    return payload

def _0083(edt):
    if edt is not None:
        if len(edt) > 1:
            ops_value = edt[1:].hex()
        else:
            ops_value = None
        return ops_value
    return None

EPC_SUPER_FUNCTIONS = {
    0x80: _0080,
    0x8A: _int,
    0x83: _0083,
    0x9E: _009X,
    0x9F: _009X
}


# ------- EPC FUNCTIONS -------------------------------------------------
# -----------------------------------------------------------------------

# ----- Tempereature Sensor -------
def _0011E0(edt):
        return float(int.from_bytes(edt, 'big')) / 10


# ----- Home Air conditioner -------
def _0130A0(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'auto',
       0x31: 'minimum',
       0x32: 'low',
       0x33: 'medium-low',
       0x34: 'medium',
       0x35: 'medium-high',
       0x36: 'high',
       0x37: 'very-high',
       0x38: 'max'
    }
    return values.get(op_mode, "Invalid setting")

# Automatic control of air flow direction setting
def _0130A1(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'auto',
       0x42: 'non-auto',
       0x43: 'auto-vert',
       0x44: 'auto-horiz'
    }
    return values.get(op_mode, "Invalid setting")

def _0130AA(edt):
    op_mode = int.from_bytes(edt, 'big')
    # print(hex(op_mode))
    values = {
      0x40: 'Normal operation',
      0x41: 'Defrosting',
      0x42: 'Preheating',
      0x43: 'Heat removal'
      }
    # return({'special':hex(op_mode)})
    return values.get(op_mode, "invalid_setting")

# Automatic swing of air flow direction setting
def _0130A3(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x31: 'not-used',
       0x41: 'vert',
       0x42: 'horiz',
       0x43: 'vert-horiz'
    }
    return values.get(op_mode, "invalid_setting")

# Air flow direction (vertical) setting
def _0130A4(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
      0x41: 'upper',
      0x44: 'upper-central',
      0x43: 'central',
      0x45: 'lower-central',
      0x42: 'lower'
      }
    # return({'special':hex(op_mode)})
    return values.get(op_mode, "invalid_setting")

# Air flow direction (horiziontal) setting
def _0130A5(edt):
    # complies with version 2.01 Release a (page 3-88)
    op_mode = int.from_bytes(edt, 'big')
    values = {
      0x41: 'rc-right',
      0x42: 'left-lc',
      0x43: 'lc-center-rc',
      0x44: 'left-lc-rc-right',
      0x51: 'right',
      0x52: 'rc',
      0x54: 'center',
      0x55: 'center-right',
      0x56: 'center-rc',
      0x57: 'center-rc-right',
      0x58: 'lc',
      0x59: 'lc-right',
      0x5A: 'lc-rc',
      0x60: 'left',
      0x61: 'left-right',
      0x62: 'left-rc',
      0x63: 'left-rc-right',
      0x64: 'left-center',
      0x65: 'left-center-right',
      0x66: 'left-center-rc',
      0x67: 'left-center-rc-right',
      0x69: 'left-lc-right',
      0x6A: 'left-lc-rc'
      }
    return values.get(op_mode, "invalid_setting")

# Operation mode
def _0130B0(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'auto',
       0x42: 'cool',
       0x43: 'heat',
       0x44: 'dry',
       0x45: 'fan_only',
       0x40: 'other'
    }
    return values.get(op_mode, "invalid_setting")

def _0260EO(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'open',
       0x42: 'close',
       0x43: 'stop'
    }
    return values.get(op_mode, "invalid_setting")

# ----- Electric Lock Class -------
def _026FEX(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'lock',
       0x42: 'unlock'
    }
    return values.get(op_mode, "invalid_setting")

def _026FE3(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'open',
       0x42: 'closed'
    }
    return values.get(op_mode, "invalid_setting")

def _026FE4(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'occupant',
       0x42: 'non-occupant'
    }
    return values.get(op_mode, "invalid_setting")

def _026FE5(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x40: 'normal',
       0x41: 'break_open',
       0x42: 'door_open',
       0x43: 'maunal_unlocked',
       0x44: 'tampered'
    }
    return values.get(op_mode, "invalid_setting")

def _026FE6(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'on',
       0x42: 'off'
    }
    return values.get(op_mode, "invalid_setting")

def _026FE7(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x41: 'normal',
       0x42: 'battery_replacement_notification'
    }
    return values.get(op_mode, "invalid_setting")

  #--- Low voltage smart meter class

def _0288E1(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
        0x00: 1,
        0x01: 0.1,
        0x02: 0.01,
        0x03: 0.001,
        0x04: 0.0001,
        0x0A: 10,
        0x0B: 100,
        0x0C: 1000,
        0x0D: 10000,
    }
    return values.get(op_mode, "invalid_setting")


# ----- Low voltage smart electric energy meter -------
def _0288E7(edt):
    value = int.from_bytes(edt, 'big', signed=True)
    return value


def _0288E8(edt):
    r_phase = float(int.from_bytes(edt[0:2], 'big', signed=True)) / 10 #R Phase
    t_phase = float(int.from_bytes(edt[2:4], 'big', signed=True)) / 10 #T Phase
    return {"r_phase_amps": r_phase, "t_phase_amps": t_phase}

def _0288EA(edt):
    print(edt)
    year = int.from_bytes(edt[0:2], 'big')
    month = int.from_bytes(edt[2:3], 'big')
    day = int.from_bytes(edt[3:4], 'big')
    hour = int.from_bytes(edt[4:5], 'big')
    minute = int.from_bytes(edt[5:6], 'big')
    second = int.from_bytes(edt[6:7], 'big')
    culmative = int.from_bytes(edt[7:], 'big')
    time = datetime(year, month, day, hour,minute,second)
    return {'time': datetime(year, month, day, hour,minute,second).isoformat(), 'culmative_value': culmative}

# ----- General lighting class -------
def _0290B1(edt):
    op_mode = int.from_bytes(edt, 'big')
    values = {
       0x40: 'other',
       0x41: 'incandescent_lamp_color',
       0x42: 'white',
       0x43: 'daylight_white',
       0x43: 'daylight_color',
    }
    return values.get(op_mode, "invalid_setting")

EPC_FUNCTIONS = {
    0x00: {
        0x11:{
            0xE0: _0011E0,
            0XE1: _int
        }
    },
    0x01: {
        0x30:{
        	0xA0: _0130A0,
            0xA1: _0130A1,
            0xA3: _0130A3,
            0xA4: _0130A4,
            0xA5: _0130A5,
            0xAA: _0130AA,
            0xB0: _0130B0,
            0xB3: _int,
            0xBB: _int,
            0xBE: _int
        }
    },
    0x02: {
        0x60: {
            0xE0: _0260EO
        },
        0x6F: {
            0xE0: _026FEX,
            0xE1: _026FEX,
            0xE2: _026FEX,
            0xE3: _026FE3,
            0xE4: _026FE4,
            0xE5: _026FE5,
            0xE6: _026FE6,
            0xE7: _026FE7
        },
        0x79: {
            0xE0: _int,
            0xE1: _int
        },
        0x7d: {
            0xE4: _int,
            0xCF: _int
        },
        0x88: {
            0xd7: _int, #number of effective digits in culmative amout of energy
            0xe0: _int, #measured culmative amount of electricity
            0xe1: _0288E1, #multiplying factor
            # 0xe2:
            0xe5: _int,
            0xe7: _0288E7, #measured instantaneous electric energy
            0xe8: _0288E8, #measured instantanenous currents
            0xea: _0288EA
        },
        0x90: {
            0xB1: _0290B1
        }
    }
}



import asyncio
import datetime

from aioudp import UDPServer
# import socket
import struct
import sys
import time
from pychonet.lib.const import GET
from pychonet.lib.functions import decodeEchonetMsg, buildEchonetMsg, preparePayload
from pychonet.lib.eojx import EOJX_GROUP, EOJX_CLASS
from pychonet.lib.const import ENL_STATUS, ENL_UID, ENL_SETMAP, ENL_GETMAP, ENL_PORT
from pychonet.lib.epc_functions import EPC_SUPER_FUNCTIONS
from pychonet.lib.epc import EPC_CODE, EPC_SUPER


class ECHONETAPIClient:
    def __init__(self, server, loop):
        self._server = server
        self._loop = loop
        self._server.subscribe(self.echonetMessageReceived)
        self._state = {}
        self._next_tx_tid = 0x01
        self._received_tids = []

    async def echonetMessageReceived(self, raw_data, addr):
        host = addr[0]
        processed_data = decodeEchonetMsg(raw_data)
        self._received_tids.append(processed_data["TID"])
        # handle discovery message response
        for opc in processed_data['OPC']:
            seojgc = processed_data['SEOJGC']
            seojcc = processed_data['SEOJCC']
            seojci = processed_data['SEOJCI']
            epc = opc['EPC']
            if seojgc == 14 and seojcc == 240 and epc == 0xd6:
                await self.process_discovery_data(host, opc)
            else:
                if epc == ENL_SETMAP:
                    map = EPC_SUPER_FUNCTIONS[epc](opc['EDT'])
                    for item in map:
                        if item not in list(self._state[host][seojgc][seojcc][seojci].keys()):
                            self._state[host][seojgc][seojcc][seojci][item] = None
                    self._state[host][seojgc][seojcc][seojci][epc] = map
                elif epc == ENL_GETMAP:
                    map = EPC_SUPER_FUNCTIONS[epc](opc['EDT'])
                    self._state[host][seojgc][seojcc][seojci][epc] = map
                elif epc == ENL_UID:
                    self._state[host][seojgc][seojcc][seojci][epc] = EPC_SUPER_FUNCTIONS[epc](opc['EDT'])
                else:
                    self._state[host][seojgc][seojcc][seojci][epc] = opc['EDT']

    async def discover(self, host = "224.0.23.0"):
        await self.echonetMessage(host, 0x0E, 0xF0, 0x00, GET, [{'EPC': 0xD6}])

    async def echonetMessage(self, host, deojgc, deojcc, deojci, esv, opc):
       if host not in list(self._state.keys()):
          self._state.update({host:{}})
       tx_tid = self._next_tx_tid
       self._next_tx_tid+=1 if self._next_tx_tid < 0xFF else 1
       payload = buildEchonetMsg({
           'TID' : tx_tid, # Transaction ID 1
           'DEOJGC': deojgc,
           'DEOJCC': deojcc,
           'DEOJCI': deojci,
           'ESV' : esv,
           'OPC' : opc
       })
       self._server.send(payload, (host, ENL_PORT))
       for x in range(0,100):
            await asyncio.sleep(0.01)
            if tx_tid in self._received_tids:
                # transaction sucessful remove from list
                self._received_tids.remove(tx_tid)
                return True
       return False

    async def getAllPropertyMaps(self, host, eojgc, eojcc, eojci):
        return await self.echonetMessage(host, eojgc, eojcc, eojci, GET, [{'EPC':ENL_GETMAP},{'EPC':ENL_SETMAP}])

    async def getIdentificationNumber(self, host, eojgc, eojcc, eojci):
        return await self.echonetMessage(host, eojgc, eojcc, eojci, GET, [{'EPC':ENL_UID}])

    async def process_discovery_data(self, host, opc_data):
            if 'discovered' not in self._state[host]:
                default_map = list(EPC_SUPER.keys())
                edt = bytearray(opc_data['EDT'])
                #1st byte: Total number of instances
                #2nd to 253rd bytes: ECHONET object codes (EOJ3 bytes) enumerated
                edtnum = bytearray(edt)[0]
                for x in range(edtnum):
                    eojgc = bytearray(edt)[1 + (3 * x)]
                    eojcc = bytearray(edt)[2 + (3 * x)]
                    eojci = bytearray(edt)[3 + (3 * x)]
                    # prepare set list
                    for epc in list(EPC_CODE[eojgc][eojcc].keys()):
                        default_map.append(epc)
                    # populate state table
                    if eojgc not in list(self._state[host].keys()):
                        self._state[host].update({eojgc:{}})
                    if eojcc not in list(self._state[host][eojgc].keys()):
                        self._state[host][eojgc].update({eojcc:{}})
                    if eojci not in list(self._state[host][eojgc][eojcc].keys()):
                        self._state[host][eojgc][eojcc][eojci] = { epc : None for epc in default_map }
                        self._state[host][eojgc][eojcc][eojci].update({ENL_SETMAP:default_map})
                        self._state[host][eojgc][eojcc][eojci].update({ENL_GETMAP:default_map})
                    await self.getIdentificationNumber(host, eojgc, eojcc, eojci)
                    if await self.getAllPropertyMaps(host, eojgc, eojcc, eojci) == True:
                        self._state[host][eojgc][eojcc][eojci]["discovered"] = True
                self._state[host]["discovered"] = True



from pychonet.EchonetInstance import EchonetInstance

class StorageBattery(EchonetInstance):

    WORKING_OPERATION_STATES = {
        0x40: "Other",
        0x41: "Rapid charging",
        0x42: "Charging",
        0x43: "Discharging",
        0x44: "Standby",
        0x45: "Test",
        0x46: "Automatic",
        0x48: "Restart",
        0x49: "Effective capacity recalculation processing"
    }

    def __init__(self, netif, instance = 0x1):
        self.eojgc = 0x02
        self.eojcc = 0x7d
        EchonetInstance.__init__(self, self.eojgc, self.eojcc, instance, netif)

    def getRemainingStoredElectricity3(self):
        return int.from_bytes(self.getSingleMessageResponse(0xE4), 'big')

    def getWorkingOperationStatus(self):
        return int.from_bytes(self.getSingleMessageResponse(0xCF), 'big')



from pychonet.echonetapiclient import ECHONETAPIClient
from .EchonetInstance import EchonetInstance
from .HomeAirConditioner import HomeAirConditioner

def Factory(instance, host, server, eojci= 0x01):

    """Factory Method"""
    instances = {
        "EchonetInstance": EchonetInstance,
        "HomeAirConditioner": HomeAirConditioner,
    }

    return instances[instance](host, server, eojci)



from pychonet.EchonetInstance import EchonetInstance

class HomeSolarPower(EchonetInstance):
    def __init__(self, netif, instance = 0x1):
        self.eojgc = 0x02
        self.eojcc = 0x79
        EchonetInstance.__init__(self, self.eojgc, self.eojcc, instance, netif)

    def getMeasuredInstantPower(self):
        return int.from_bytes(self.getSingleMessageResponse(0xE0), 'big')

    def getMeasuredCumulPower(self):
        return int.from_bytes(self.getSingleMessageResponse(0xE1), 'big')



from .eojx import EOJX_GROUP, EOJX_CLASS
from .const import ESV_CODES, EHD1, EHD2

def decodeEchonetMsg(byte):
  data = {}
  try:
      data['EHD1'] = byte[0]
      if data['EHD1'] not in EHD1:
          raise ValueError('EHD1 Header invalid')
      data['EHD2'] = byte[1]
      if data['EHD2'] not in EHD2:
          raise ValueError('EHD2 Header invalid')
      data['TID'] = int.from_bytes(byte[2:4], byteorder='big')

      # Decode SEOJ
      data['SEOJGC'] = byte[4]
      data['SEOJCC'] = byte[5]
      data['SEOJCI'] = byte[6]
      # Decode DEOJ
      data['DEOJGC'] =  byte[7]
      data['DEOJCC'] =  byte[8]
      data['DEOJCI'] =  byte[9]

      # Decode Service property
      data['ESV'] =  byte[10]

      i = 0
      epc_pointer = 12
      data['OPC'] = []
      # decode multiple processing properties (OPC)
      while i <  (byte[11]):
          OPC = {}
          pdc_pointer = epc_pointer + 1
          edt_pointer = pdc_pointer + 1
          end_pointer = edt_pointer
          OPC['EPC'] = byte[epc_pointer]
          OPC['PDC'] =  byte[pdc_pointer]
          pdc_len = byte[pdc_pointer]
          end_pointer += pdc_len
          OPC['EDT'] = byte[edt_pointer:end_pointer]
          epc_pointer = end_pointer
          i += 1
          data['OPC'].append(OPC)

  except ValueError as error:
        print('Caught this error: ' + repr(error))
        quit()
  return data

def buildEchonetMsg(data):
   try:
      # EHD is fixed to 0x1081 because I am lazy.
      message = 0x1081

      # validate TID (set a default value if none provided)
      # TODO - TID message overlap.
      if 'TID' not in data:
         data['TID'] = 0x01
      elif data['TID'] > 0xFFFF:
         raise ValueError('Transaction ID is larger then 2 bytes.')
      message = (message << 16) + data['TID']

      # append SEOJ which again is a fixed value to be lazy
      message = (message << 24) + 0x05FF01

      # validate DEOJ
      if data['DEOJGC'] in EOJX_GROUP:
          message = (message << 8) + data['DEOJGC']
      else:
          raise ValueError('Value ' + str(hex(data['DEOJGC'])) + ' not a valid SEO Group code')

      if data['DEOJCC'] in EOJX_CLASS[data['DEOJGC']]:
          message = (message << 8) + data['DEOJCC']
      else:
          raise ValueError('Value ' + str(hex(data['DEOJCC'])) + ' not a valid SEO class code')

      message = (message << 8) + data['DEOJCI']

      # validate ESV by looking up the codes.
      if data['ESV'] in ESV_CODES:
          # ESV code is a string
          message = (message << 8) + data['ESV']
      else:
          raise ValueError('Value not in ESV code table')

      # validate OPC
      message = (message << 8) + len(data['OPC'])

      # You can have multiple OPC per transaction.
      for values in data['OPC']:
        # validate EPC
        message =  (message << 8) + values['EPC']
        if 'PDC' in values:
            message =  (message << 8) + values['PDC']
        # if PDC has a value then concat EDT to message
            if values['PDC'] > 0:
                message =  (message << 8 * values['PDC']) + values['EDT']
        else:
            message =  (message << 8) + 0x00
      # print(message)
      return bytearray.fromhex(format(int(message), 'x'))
# some sloppy error handling here.
   except ValueError as error:
        # print('Caught this error: ' + repr(error))
        quit()

def preparePayload(tid, deojgc, deojcc, deojci, esv, opc):
   tx_payload = {
       'TID' : tid, # Transaction ID 1
       'DEOJGC': deojgc,
       'DEOJCC': deojcc,
       'DEOJCI': deojci,
       'ESV' : esv,
       'OPC' : opc
   }
   return tx_payload



import asyncio
from pychonet.echonetapiclient import ECHONETAPIClient
from pychonet.lib.const import GET, SETC, ENL_SETMAP, ENL_GETMAP, ENL_OFF, ENL_ON, ENL_STATUS, ENL_UID
from pychonet.lib.epc_functions import EPC_SUPER_FUNCTIONS
from pychonet.lib.epc import EPC_CODE, EPC_SUPER



"""
Superclass for Echonet instance objects.
"""
class EchonetInstance:

    EPC_FUNCTIONS = {}
    """
    Constructs an object to represent an Echonet lite instance .

    :param eojgc: Echonet group code
    :param eojcc: Echonet class code
    :param instance: Instance ID
    :param netif: IP address of node
    """
    def __init__(self, host, eojgc, eojcc, instance, api_connector=None):
        self._host = host
        self._eojgc = eojgc
        self._eojcc = eojcc
        self._eojci = instance
        self._api = api_connector

    """
    getMessage is used to fire a single ECHONET get messages to get Node information
    Assumes one EPC is sent per message.

    :param tx_epc: EPC byte code for the request.
    :return: the deconstructed payload for the response

    """
    async def getMessage(self, epc, pdc = 0x00):
        opc = [{'EPC': epc, 'PDC': pdc}]
        response = await self._api.echonetMessage(self._host, self._eojgc, self._eojcc, self._eojci, GET, opc)
        if not response:
            return False
        edt = self._api._state[self._host][self._eojgc][self._eojcc][self._eojci][epc]
        if epc in list(EPC_SUPER_FUNCTIONS.keys()):
            return EPC_SUPER_FUNCTIONS[epc](edt)
        return edt

    """
    setMessage is used to fire ECHONET set messages to set Node information
    Assumes one OPC is sent per message.

    :param tx_epc: EPC byte code for the request.
    :param tx_edt: EDT data relevant to the request.
    :return: True if sucessful, false if request message failed
    """
    async def setMessage(self, epc, edt, pdc = 0x01):
        opc = [{'EPC': epc, 'PDC': pdc, 'EDT': edt}]
        response = await self._api.echonetMessage(self._host, self._eojgc, self._eojcc, self._eojci, SETC, opc)
        if not response:
            return False
        return True

    """
    setMessage is used to fire ECHONET set messages to set Node information
    Assumes one OPC is sent per message.

    :param tx_epc: EPC byte code for the request.
    :param tx_edt: EDT data relevant to the request.
    :return: True if sucessful, false if request message failed
    """
    async def setMessages(self, opc):
        response = await self._api.echonetMessage(self._host, self._eojgc, self._eojcc, self._eojci, SETC, opc)
        if not response:
            return False
        return True

    """
    update is used as a way of producing a dict useful for API polling etc
    Data can be formatted or returned as a hex string value by default.

    :param attributes: optional list of EPC codes. eg [0x80, 0xBF], or a single code eg 0x80

    :return dict: A dict with the following attributes:
    {128: 'On', 160: 'medium-high', 176: 'heat', 129: '00', 130: '00004300',
    131: '0000060104a0c9a0fffe069719013001', 179: 19, 134: '06000006000000020000', 136: '42', 137: '0000'}

    :return string: if attribute is a single code then return value directly:
    eg:
    update(0x80)
    'on'

    """
    async def update(self, attributes=None):
        opc = []
        if attributes == None:
            attributes = self.getGetProperties()
        if isinstance(attributes, int):
            list_attributes = [attributes]
            attributes = list_attributes
        returned_json_data = {}
        for value in attributes:
          if value in self.getGetProperties():
            opc.append({'EPC': value})
        response = await self._api.echonetMessage(self._host, self._eojgc, self._eojcc, self._eojci, GET, opc)
        if response is not False:
             for epc in attributes:
                 if epc not in list(self._api._state[self._host][self._eojgc][self._eojcc][self._eojci].keys()):
                     returned_json_data.update({epc: False})
                     continue
                 elif epc in list(EPC_SUPER_FUNCTIONS.keys()): # check if function is defined in the superset
                     returned_json_data.update({epc: EPC_SUPER_FUNCTIONS[epc](self._api._state[self._host][self._eojgc][self._eojcc][self._eojci][epc])})
                     continue
                 elif epc in list(EPC_SUPER.keys()): # return hex value if code exists in superset but no function found
                     returned_json_data.update({epc: self._api._state[self._host][self._eojgc][self._eojcc][self._eojci][epc].hex()})
                     continue
                 elif epc in list(self.EPC_FUNCTIONS.keys()):
                     returned_json_data.update({epc: self.EPC_FUNCTIONS[epc](self._api._state[self._host][self._eojgc][self._eojcc][self._eojci][epc])})
                     continue
                 elif epc in list(EPC_CODE[self._eojgc][self._eojcc].keys()): # return hex value if EPC code exists in class but no function found
                     returned_json_data.update({epc: self._api._state[self._host][self._eojgc][self._eojcc][self._eojci][epc].hex()})
        for epc in attributes:
             if epc not in list(returned_json_data.keys()):
                 returned_json_data.update({epc: False})
        if(len(returned_json_data)) == 1 and len(attributes) == 1:
             return returned_json_data[attributes[0]]
        elif(len(returned_json_data)) == 0:
             return False
        return returned_json_data

    """
    getIdentificationNumber returns a number used to identify an object uniquely

    :return: Identification number as a string.
    """
    async def getIdentificationNumber(self): # EPC 0x83
        await self._api.getIdentificationNumber(self._host, self._eojgc, self._eojcc, self._eojci)
        return self._api._status[self._host][self._eojgc][self._eojcc][self._eojci][ENL_UID]

    """
    getOperationalStatus returns the ON/OFF state of the node

    :return: status as a string.
    """
    async def getOperationalStatus(self): # EPC 0x80
        return await self.getMessage(ENL_STATUS)

    """
    setOperationalStatus sets the ON/OFF state of the node

    :param status: True if On, False if Off.
    """
    async def setOperationalStatus(self, status):
        if status == 'on' or status == ENL_ON or status == True:
            return await self.setMessage(ENL_STATUS, ENL_ON)
        elif status == 'of' or status == ENL_OFF or status == False:
            return await self.setMessage(ENL_STATUS, ENL_OFF)

    """
    On sets the node to ON.

    """
    async def on (self): # EPC 0x80
        return await self.setMessage(ENL_STATUS, ENL_ON)

    """
    Off sets the node to OFF.

    """
    async def off (self): # EPC 0x80
        return await self.setMessage(ENL_STATUS, ENL_OFF)

    def getSetProperties (self): # EPC 0x9E
        return self._api._state[self._host][self._eojgc][self._eojcc][self._eojci][ENL_SETMAP]

    def getGetProperties (self): # EPC 0x9F
        return self._api._state[self._host][self._eojgc][self._eojcc][self._eojci][ENL_GETMAP]

    async def getAllPropertyMaps(self):
        return {ENL_SETMAP: self.getSetProperties(), ENL_GETMAP: self.getGetProperties()}
